# -*- coding: utf-8 -*-
#
# This file is part of fedscm_admin.
# Copyright © 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
"""
Provides helper functions for FAS
"""
from __future__ import absolute_import

import fasjson_client
from fedora.client import AuthError
from click import ClickException

from fedscm_admin import CONFIG
from fedscm_admin.config import get_config_item


class FASClient(object):
    """
    A helper class to maintain a FAS session
    """
    def __init__(self):

        self.fasjson = get_config_item(CONFIG, 'fasjson')
        if self.fasjson:
            fasjson_url = get_config_item(CONFIG, 'fasjson_url')
            try:
                self.client = fasjson_client.Client(url=fasjson_url)
            except fasjson_client.errors.ClientSetupError:
                raise ClickException('Failed to create fasjson_client.')
        else:
            self.unauthenticated_error = (
                'The FAS Client is not authenticated. Please make sure you typed '
                'in the correct credentials.')

    def set_credentials(self, username, password):
        """
        A login function to FAS. If the password is incorrect, a ClickException
        is raised.
        :param username: a string of the username
        :param password: a string of the password
        :return: None
        """
        if not self.client.verify_password(username, password):
            raise ClickException(
                'The login to FAS failed. Please make sure you typed in the '
                'correct credentials.')
        self.client.username = username
        self.client.password = password

    def set_fasjson_username(self, username):
        """
        A simple function to place the username on the client.
        This is to preserve backwards compatibility only, the username
        associated with fasjson isn't used anywhere
        :param username: a string of the username
        :param password: a string of the password
        :return: None
        """
        self.client.username = username

    def get_fas_user_by_id(self, user_id):
        """
        Get the FAS user based on the user ID
        :param user_id: an integer of the FAS user ID
        """
        try:
            return self.client.person_by_id(user_id)
        except AuthError:  # pragma: no cover
            raise ClickException(self.unauthenticated_error)

    def get_fas_user(self, value, search_key='username'):
        """
        Taken from:
        https://pagure.io/pkgdb-cli/blob/main/f/pkgdb2client/utils.py

        Get the FAS user associated with the provided key value
        If an email address is used, it can be either the FAS email address or
        the one used in Bugzilla.
        :param value: a string to search for
        :param search_key: a string of the key to search for. This defaults to
        email
        :return: a FAS user or None
        """

        if self.fasjson:
            if search_key == 'email':
                email = value
                try:
                    res = self.client.search(email=email).result
                    if res:
                        return res[0]
                    else:
                        return None
                except fasjson_client.errors.APIError as e:
                    raise ClickException(
                            f"Encountered an error trying to retrieve user. {e.message}"
                    )
            else:
                try:
                    return self.client.get_user(username=value).result
                except fasjson_client.errors.APIError as e:
                    if e.code == 404:
                        return None
                    else:
                        raise ClickException(
                            f"Encountered an error trying to retrieve user. {e.message}"
                        )
        else:
            email = None
            if search_key == 'email':
                email = value
            if email and email in self.client._AccountSystem__alternate_email:
                user_id = self.client._AccountSystem__alternate_email[email]
            else:
                try:
                    user_id = self.client.people_query(
                        constraints={search_key: value},
                        columns=['id']
                    )
                except AuthError:  # pragma: no cover
                    raise ClickException(self.unauthenticated_error)
                if user_id:
                    user_id = user_id[0].id

            if user_id:
                return self.get_fas_user_by_id(user_id)
        return None

    def user_member_of(self, user, group):
        """
        Checks
        :param user: a FAS user object
        :param group: a string of the FAS group
        :return: boolean
        """
        if self.fasjson:
            try:
                return self.client.check_membership(groupname=group,
                                                    username=user['username']).result
            except fasjson_client.errors.APIError as e:
                raise ClickException(f"Encountered an error trying to check membership. \
                    {e.message}")
        else:
            return user and group in user['group_roles'] \
                and user['group_roles'][group]['role_status'] == 'approved'
